/*
*************************************************************************************************
*****																							*
*****						GESCHWINDIGKEITS-  UND  ZEITMESSGERT								*
*****																							*
*****							09.2005 - 09.2005  Thilo Mlls									*
*****																							*
*****	FUNKTION:																				*
*****	---------																				*
*****	Die Zeitmessung wird mit LOW-Flanke an INT0 (Pin 16) gestartet und mit LOW-Flanke an	*
*****	INT1 (Pin 17) gestoppt. Die Auflsung betrgt 0.5s und der Zeitbereich geht bis		*
*****	1 * 10E300 Tage. Als Prozessor wurde ein ATmega32 16PI mit 16MHz Taktfrequenz verwendet.*
*****	Zeit- und Geschwindigkeit werden auf RS232 mit 19200,n,8,1 Baud ausgegeben.				*
*****	Der Anzeigentreiber fr ein 2 x 40 Zeilen-LCD-Display wurde eingebunden.				*
*****	Der Abstand der Lichtschranken wird in der Variablen 'm_Strecke_d' im double-Format		*
*****	(hier 4.7cm) festgelegt.																*
*****	Die Genauigkeit hngt nur von der Genauigkeit des Quarzes ab. Die Interruptprioritten	*
*****	0 und 1 sind am hchsten angesiedelt und verzgern die Messung deshalb auf keine Weise.	*
*****																							*
*****	Zur optischen kontrolle dass die Zeit luft wurde der Ausgang PORTD7 (Pin 21) zur		*
*****	Ansteuerung einer LED verwendet (LED ein => Zeit luft).								*
*****																							*
*****		 																					*
*****		 																					*
*****		 																					*
*****	_______________________________________________________________________________________	*
*****	Quellcode fr GNU-C-Compiler (AVR-GCC)													*
*************************************************************************************************
_________________________________________________________________________________________________
| Belegung der Ein-/ Ausgnge:																	|
| ----------------------------																	|
| PORT A: 	Bit 0: Pin 40: (ADC0) - NC -														|	
| -------	Bit 1: Pin 39: (ADC1) - NC -														|
|			Bit 2: Pin 38: (ADC2) - NC -														|
|			Bit 3: Pin 37: (ADC3) - NC -														|
|			Bit 4: Pin 36: (ADC4) - NC -														|
|			Bit 5: Pin 35: (ADC5) - NC -														|
|			Bit 6: Pin 34: (ADC6) - NC -														|
|			Bit 7: Pin 33: (ADC7) - NC -														|
| PORT B: 	Bit 0: Pin 01: (XCK/T0) - NC -														|
| -------	Bit 1: Pin 02: (T1) - NC -															|
|			Bit 2: Pin 03: (INT2/AIN0) - NC -													|
|			Bit 3: Pin 04: (OC0/AIN1) - NC -													|
|			Bit 4: Pin 05: (NSS) - NC -															|
|			Bit 5: Pin 06: (MOSI) ISP															|
|			Bit 6: Pin 07: (MISO) ISP															|
|			Bit 7: Pin 08: (SCK) ISP															|
| PORT C: 	Bit 0: Pin 22: (SCL) - NC -															|
| -------	Bit 1: Pin 23: (SDA) - NC -															|
|			Bit 2: Pin 24: (TCK) - NC -															|
|			Bit 3: Pin 25: (TMS) - NC -															|
|			Bit 4: Pin 26: (TDO) - NC -															|
|			Bit 5: Pin 27: (TDI) - NC -															|
|			Bit 6: Pin 28: (TOSC1) - NC -														|
|			Bit 7: Pin 29: (TOSC2) - NC -														|
| PORT D: 	Bit 0: Pin 14: (RXD) RS232															|
| -------	Bit 1: Pin 15: (TXD) RS232															|
|			Bit 2: Pin 16: (INT0) Zeit-Startsignal (Low-Impulse)								|
|			Bit 3: Pin 17: (INT1) Zeit-Stopsignal (Low-Impulse)									|
|			Bit 4: Pin 18: (OC1B) - NC -														|
|			Bit 5: Pin 19: (OC1A) - NC -														|
|			Bit 6: Pin 20: (ICP) - NC -															|
|			Bit 7: Pin 21: (OC2) LED "Zeit luft"												|
|________________________________________________________________________________________________
*/
#include <avr/signal.h>
#include <avr/io.h>
#include <stdio.h>
#include <inttypes.h>
#include <stdlib.h>
#define SYSCLK 16000000						// Prozessortakt festlegen
#include <avr/delay.h>
#include <compat/twi.h>
#include <Display16x2.c>					// Sourcecode fr Displaysteuerung
#include <avr/eeprom.h>

// Bitdefinitionen 

unsigned char TWI_NWR = 1;		// Wrie/Read - Umschaltung


//  Globale Variablen im SRAM

volatile unsigned char STATUS = 0;			// Statusvariable
volatile unsigned char Temp = 0;				// Tepmporr-Variable
unsigned char R_Byte = 0;
volatile double us_Zaehler = 0;				// s - Zhler
volatile double ms_Zaehler = 0;				// ms - Zhler
volatile double s_Zaehler = 0;				// Sekundenzhler
volatile double min_Zaehler = 0;				// Minutenzhler
volatile double h_Zaehler = 0;				// Stundenzhler
volatile double d_Zaehler = 0;				// Tageszhler
volatile double kmh = 0;						// km/h - Zhler
volatile double Sekunden = 0;					// Sekunden fr Geschwindigkeitsberechnung
volatile double m_Strecke_d = 0;				// Lnge der Messstrecke in m
volatile unsigned char SLA_W = 0;			// TWI-Adresse (senden)
volatile unsigned char DATA = 0;				// TWI-Daten (senden)
volatile unsigned char DEV_SEL = 0;			// TWI-Device-Selector (senden)
volatile unsigned char DEV_ADR = 0;			// TWI-Device-Adresse (senden)


int USART_Transmit(char txtemp)				// RS232-Senderoutine
{
	while ( !( UCSRA & (1<<UDRE))); 			// warten, bis Sendepuffer leer ist
	UDR = txtemp;								// Byte senden
	return 0;
}

int main (void)								// Start
{	

// Datenrichtung der Ports

	DDRC &= ~(1<<PC0) & ~(1<<PC1);
	PORTC |= (1<<PC1)|(1<<PC0);	
	DDRD &= ~(1<<PD2) & ~(1<<PD3);	// Int0 und INT1 auf Eingang
	DDRD |= (1<<PD7);					// LED Ausgang
	PORTD &= ~(1<<PD2) & ~(1<<PD3);	// kein Pullup fr Int0 und INT1
	
// UART Einstellungen

	UCSRA = 0b00000000;
	UCSRB = 0b10011000;					// RX & TX enable, RX-Int. enabled
	UBRRL = 51;							// 19200 Baud bei 16 MHz, 0.2% Fehler
	
// TWI Einstellungen

	TWSR = 0b000000000;					// Prescaler /64 => 250kBaud
	//TWCR |= (1<<TWIE);					// TWI enabled
	TWAR = 0;							// Slave-Adresase auf 0
	TWBR = (F_CPU / 100000UL - 16) / 2;
	
// Einstellungen fr Timer 1

	TCCR1A = 0b00000000;
	TCCR1B = 0b00001000;				// Prescaler /8, CTC enabled (TCCR1B = 0b00001010;=> einschalten)
	TIMSK = 0b00010000;					// Output-compare1-ovf enabled
	TCNT1 = 0;							// Timer rcksetzen
	
	OCR1A = 1999;		// 1ms, kalibriert		// OCR1A = 2000 (16MHz / 8 / 2000 = 1ms)

// Einstellungen fr Interrupts und Sleepmode
	
	MCUCR |= (1<<ISC01)|(1<<ISC11);		// INT0 und INT1 falling edge
	MCUCR &= ~(1<<ISC00) & ~(1<<ISC10);
	GIFR |= (1<<INTF0)|(1<<INTF1);			// INT-Flags lschen
	GICR |= (1<<INT0);						// INT0 enabled
	GICR &= ~(1<<INT1);					// INT1 disabled

	SREG |= (1<<SREG_I);					// Globalen Interrupt freigeben

	m_Strecke_d = 0.2;					// Messabstand 20cm
	
fclose (stdout);					// Output-Stream schlieen
fdevopen(USART_Transmit, NULL, 0);	// Output-Stream auf RS232-Ausgabe
	
void ERROR (void)
	{
	}
	
void TWI_read (void)
{

	DEV_SEL = 0b10100001;	
start:
	TWCR |= (1<<TWINT)|(1<<TWSTA)|(1<<TWEN);		// Startbedingung senden
	while ((!(TWCR & (1<<TWINT))));				// Warten bis TWINT-Flag gesetzt ist
	
	TWDR = DEV_SEL;								// Adresse ins Datenregister laden
	TWCR |= (1<<TWINT) | (1<<TWEN);			// Adresse senden
	while ((!(TWCR & (1<<TWINT))));				// warten bis Adresse gesendet wurde

	TWCR |= (1<<TWINT)|(1<<TWEN)|(1<<TWSTO);	// Stopbedingung senden
	while ((!(TWCR & (1<<TWINT))));				// Warten bis TWINT-Flag gesetzt ist
	

	DATA = TWDR;
}
	
	
for (;;)		// Endlosschleife main
	{
	if (STATUS == 1)
	{
	STATUS = 0;
	for (DEV_ADR=0;DEV_ADR<=16;DEV_ADR++)
		{
			TWI_read();
			printf ("%c",DATA);				// Messwerte ausgeben
		}
	}
	}

}
/***************************************************************************************/
/*** Ende main *************************************************************************/
/***************************************************************************************/

SIGNAL (SIG_UART_RECV)			// RX Interrupt-complete-Routine
{
	R_Byte = UDR;				// Empfangenes Byte in Globale Variable einlesen
}

/***************************************************************************************/

SIGNAL (SIG_OUTPUT_COMPARE1A) 	// ms-Zhler
{
	ms_Zaehler ++;
	
	if (ms_Zaehler >= 1000)
		{
			s_Zaehler ++;		// ms hochzhlen
			ms_Zaehler = 0;
		}
	if (s_Zaehler >= 60)
		{
			min_Zaehler ++;		// Minuten hochzhlen
			s_Zaehler = 0;
		}
	if (min_Zaehler >= 60)
		{
			h_Zaehler ++;		// Stunden hochzhlen
			min_Zaehler = 0;
		}
	if (h_Zaehler >= 24)
		{
			d_Zaehler ++;		// Tage hochzhlen
			h_Zaehler = 0;
		}
}

/***************************************************************************************/

SIGNAL (SIG_INTERRUPT0)			// Zeitmessung starten
{
	TCCR1B = 0b00001010;				// Timer starten
	PORTD |= (1<<PD7);					// LED ein
	
	// zur Entprellung nur den komplementren Interrupt freigeben
	GIFR |= (1<<INTF0)|(1<<INTF1);		// INT-Flags lschen
	GICR |= (1<<INT1);					// INT1 enabled
	GICR &= ~(1<<INT0);				// INT0 disabled
	STATUS = 1;
}

/***************************************************************************************/

SIGNAL (SIG_INTERRUPT1)			// Zeitmessung stoppen
{
	TCCR1B = 0b00001000;			// Timer stoppen
	PORTD &= ~(1<<PD7);			// LED aus
	
		us_Zaehler = TCNT1;
		us_Zaehler = (us_Zaehler)/2000;		// s berechnen (auf 1/2000 genau)
		Sekunden = (h_Zaehler*3600) + (min_Zaehler*60) + (s_Zaehler) + (ms_Zaehler/1000) + (us_Zaehler/1000);
		kmh = (m_Strecke_d/Sekunden)*3.6;	//auf km/h umrechnen
		printf ("Zeit: %02g.%02g:%02g:%02g.%03g.%03g => % 4.3g km/h\r\n",d_Zaehler,h_Zaehler,min_Zaehler,s_Zaehler,ms_Zaehler,us_Zaehler*1000,kmh);// Messwerte ausgeben

	// Zhler zurcksetzen
	us_Zaehler = 0;		// s - Zhler
	ms_Zaehler = 0;		// ms - Zhler
	s_Zaehler = 0;		// Sekundenzhler
	min_Zaehler = 0;	// Minutenzhler
	h_Zaehler = 0;		// Stundenzhler
	d_Zaehler = 0;		// Tageszhler
	kmh = 0;			// km/h - Zhler
	TCNT1 = 0;			// Zhler rcksetzen

	// zur Entprellung nur den komplementren Interrupt freigeben
	GIFR |= (1<<INTF0)|(1<<INTF1);		// INT-Flags lschen
	GICR |= (1<<INT0);					// INT0 enabled
	GICR &= ~(1<<INT1);				// INT1 disabled
}

/***************************************************************************************/

SIGNAL (SIG_2WIRE_SERIAL)		// IC - Empfangsinterrupt
{
DATA = TWDR;
printf ("%c",DATA);				// Messwerte ausgeben
}

/***************************************************************************************/



